/*
STAGE ONE: Load the text file containing the level data.
*/
#include <iostream>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

using namespace std;

char buffer[2048];
unsigned char mapFormat[2128], romData[262144];
int mapFormatID = 0, mapSize;

void message(char errorMessage[])
{
    cout << errorMessage << "\nPress enter to continue.";
    cin.ignore();
    return;

}

//Declare file-scanning pointers and buffer
FILE *romLoad_ptr;
FILE *romSave_ptr;
FILE *map_ptr;
FILE *text_ptr;

//Include the supported file formats for conversions
#include "Map Types/MappyMap.h"
#include "Map Types/SMSMap.h"

//Include the traps header file
#include "Level Data/Traps.h"

//ARRAY_COUNTER is used to find out how many items there are in the STAT_NUMBER array. Make sure it is listed last.
enum    {STAT_ROM_LOAD,
         STAT_ROM_SAVE,
         STAT_MAP_FILE,
         STAT_LEVEL,
         STAT_RATING,
         STAT_TILESET,
         STAT_NAME,
         STAT_LEMMINGS,
         STAT_TOBESAVED,
         STAT_RELEASERATE,
         STAT_TIME,
         STAT_CLIMBERS,
         STAT_FLOATERS,
         STAT_BOMBERS,
         STAT_BLOCKERS,
         STAT_BUILDERS,
         STAT_BASHERS,
         STAT_MINERS,
         STAT_DIGGERS,
         STAT_FALLDISTANCE,
         STAT_TRAPTYPE,
         STAT_TRAPX,
         STAT_TRAPY,
         ARRAY_COUNTER};

const char *STAT_NUMBER[] = {  "ROM LOAD",
                               "ROM SAVE",
                               "MAP FILE",
                               "LEVEL",
                               "RATING",
                               "TILESET",
                               "NAME",
                               "LEMMINGS",
                               "TO BE SAVED",
                               "RELEASE RATE",
                               "TIME",
                               "CLIMBERS",
                               "FLOATERS",
                               "BOMBERS",
                               "BLOCKERS",
                               "BUILDERS",
                               "BASHERS",
                               "MINERS",
                               "DIGGERS",
                               "FALL DISTANCE",
                               "TRAP TYPE",
                               "TRAP X",
                               "TRAP Y"};

int statNumber (char statName[])
{
    for (int j = 0; j != ARRAY_COUNTER; j++)
    {
        if (strcmp(statName, STAT_NUMBER[j]) == 0)
        {
            return j;
        }
    }
    return -1;
}

int stringToNumber (int offset)
{
    int length = 0, j = 1;
    while (j != 0)
    {
        if ((buffer[offset+length] >= '0') && (buffer[offset+length] <= '9'))
        {
            length++;
        }
        else
        {
            j = 0;
        }
    }

    if (length == 0)
    {
        return 0;
    }

    char number[length+1];
    strncpy (number,buffer+offset,length);
    number[length] = '\0';
    return atoi(number);
}

int stringLength(int offset)
{
    if (buffer[offset] != '"')
    {
        return 0;
    }

    int length = 0;
    char letter;
    while (letter != '"')
    {
        length++;
        letter = buffer[offset+length];
    }

    return (length - 1);
}

int main()
{


    int num;

    //Declare file name holders
    char romLoad[256] = "\0", romSave[256] = "\0", map[256] = "\0";

    //Declare level stats
    char levelName[32] = "\0";
    int mapFileType = 0, level = 0, rating = 0, tileset = -1, lemmings = 0, toBeSaved = 0, releaseRate = 1, timeMinutes = 5, climbers = 0, floaters = 0, bombers = 0, blockers = 0, builders = 0, bashers = 0, miners = 0, diggers = 0, fallDistance = 56, trapType = 0, trapX = 0, trapY = 0;

    //Load leveldata.txt
    text_ptr = fopen("leveldata.txt", "r");

    //Check if the file loaded
    if (text_ptr == NULL)
    {
        message("Unable to open leveldata.txt");
        return 1;
    }

    //Offload the text file into the buffer
    num = fread (buffer, 1, 1000, text_ptr);

    //Free the file fromLoad memory
    fclose (text_ptr);

    //Acquire all the variables fromLoad the buffer
    //openBracket and closeBracket are the locations of the < and > inside the text file
    int i = -1, openBracket[128], closeBracket[128], bracketCount = 0;
    bool caseToggle = 0;

    //Scanning through the file
    while (++i < num)
    {
        //Converts letter to upper case
        if (buffer[i] == '<')
        {
            caseToggle = 1;
        }
        if (buffer[i] == '>')
        {
            caseToggle = 0;
        }
        if ((buffer[i] >= 'a') && (buffer[i] <= 'z') && caseToggle)
        {
            buffer[i] -= 32;
        }

        //Detects <
        if (buffer[i] == '<')
        {
            openBracket[bracketCount] = i;
        }

        //Detects >, and increases bracketCount
        if (buffer[i] == '>')
        {
            closeBracket[bracketCount++] = i;
        }
    }

    //Declaring string-holding variable
    char readString[262];
    int strLength;

    //Looks through the file again, now assigning the level stat variables this time
    for (i = 0; i < bracketCount; i++)
    {
        //num is used to calculate the length of the word between the <>'s
        num = closeBracket[i] - openBracket[i] - 1;

        //Read the word between the <>'s
        strncpy (readString, buffer + (openBracket[i] + 1), num);
        readString[num] = '\0';

        //Performs a task based on the detected word
        switch(statNumber(readString))
        {
            case STAT_ROM_LOAD :
            {
                strLength = stringLength(closeBracket[i]+2);
                strncpy (romLoad,buffer + (closeBracket[i] + 3), strLength);
                romLoad[strLength] = '\0';
                //cout << romLoad;
                break;
            }
            case STAT_ROM_SAVE :
            {
                strLength = stringLength(closeBracket[i]+2);
                strncpy (romSave,buffer + (closeBracket[i] + 3), strLength);
                romSave[strLength] = '\0';
                //cout << romSave;
                break;
            }
            case STAT_MAP_FILE :
            {
                strLength = stringLength(closeBracket[i]+2);
                strncpy (map,buffer + (closeBracket[i] + 3), strLength);
                map[strLength] = '\0';

                //Lowers the case of the file type for comparison purposes
                for (int mapCase = strLength - 4; mapCase <= strLength; mapCase++)
                {
                    if ((map[mapCase] >= 'A') && (map[mapCase] <= 'Z'))
                        {
                            map[mapCase] += 32;
                        }
                }

                //Sets the value based on the file type. 1 = .MLM, 2 = .MAP.
                if (strcmp(map + (strLength-4), ".mlm") == 0) mapFileType = 1;
                if (strcmp(map + (strLength-4), ".map") == 0) mapFileType = 2;
                //cout << map;
                break;
            }
            case STAT_LEVEL :
            {
                level = stringToNumber(closeBracket[i]+2);
                //cout << level;
                level--;
                break;
            }
            case STAT_RATING :
            {
                rating = stringToNumber(closeBracket[i]+2);
                //cout << rating;
                rating--;
                break;
            }
            case STAT_NAME :
            {
                strLength = stringLength(closeBracket[i]+2);
                strncpy (levelName,buffer + (closeBracket[i] + 3), strLength);
                levelName[strLength] = '\0';

                //The level name MUST be stored in upper case, otherwise it won't be rendered properly.
                for (int i=0; i<strLength; i++)
                {
                    if ((levelName[i] >= 'a') && (levelName[i] <= 'z'))
                        {
                            levelName[i] -= 32;
                        }
                }
                //cout << levelName;
                break;
            }
            case STAT_TILESET :
            {
                tileset = stringToNumber(closeBracket[i]+2);
                //cout << tileset;
                break;
            }
            case STAT_LEMMINGS :
            {
                lemmings = stringToNumber(closeBracket[i]+2);
                //cout << lemmings;
                break;
            }
            case STAT_TOBESAVED :
            {
                toBeSaved = stringToNumber(closeBracket[i]+2);
                //cout << toBeSaved;
                break;
            }
            case STAT_RELEASERATE :
            {
                releaseRate = stringToNumber(closeBracket[i]+2);
                //cout << releaseRate;
                break;
            }
            case STAT_TIME :
            {
                timeMinutes = stringToNumber(closeBracket[i]+2);
                //cout << timeMinutes;
                break;
            }
            case STAT_CLIMBERS :
            {
                climbers = stringToNumber(closeBracket[i]+2);
                //cout << climbers;
                break;
            }
            case STAT_FLOATERS :
            {
                floaters = stringToNumber(closeBracket[i]+2);
                //cout << floaters;
                break;
            }
            case STAT_BOMBERS :
            {
                bombers = stringToNumber(closeBracket[i]+2);
                //cout << bombers;
                break;
            }
            case STAT_BLOCKERS :
            {
                blockers = stringToNumber(closeBracket[i]+2);
                //cout << blockers;
                break;
            }
            case STAT_BUILDERS :
            {
                builders = stringToNumber(closeBracket[i]+2);
                //cout << builders;
                break;
            }
            case STAT_BASHERS :
            {
                bashers = stringToNumber(closeBracket[i]+2);
                //cout << bashers;
                break;
            }
            case STAT_MINERS :
            {
                miners = stringToNumber(closeBracket[i]+2);
                //cout << miners;
                break;
            }
            case STAT_DIGGERS :
            {
                diggers = stringToNumber(closeBracket[i]+2);
                //cout << diggers;
                break;
            }
            case STAT_FALLDISTANCE :
            {
                fallDistance = stringToNumber(closeBracket[i]+2);
                //cout << fallDistance;
                break;
            }
            case STAT_TRAPTYPE :
            {
                trapType = stringToNumber(closeBracket[i]+2);
                break;
            }
            case STAT_TRAPX :
            {
                //An 'x' is used to override the trap X calculation.
                if (((buffer[closeBracket[i]+2]) == 'x') || (buffer[closeBracket[i]+2]) == 'X')
                {
                    trapX = stringToNumber(closeBracket[i]+3);
                }
                else
                {
                    trapX = stringToNumber(closeBracket[i]+2) * 8;
                    if ((trapType == 2) || (trapType == 4))
                    {
                        trapX += 4;
                    }
                }
                break;
            }
            case STAT_TRAPY :
            {
                //A 'y' is used to override the trap Y calculation.
                if (((buffer[closeBracket[i]+2]) == 'y') || (buffer[closeBracket[i]+2]) == 'Y')
                {
                    trapY = stringToNumber(closeBracket[i]+3);
                }
                else
                {
                    trapY = (stringToNumber(closeBracket[i]+2) * 8)+8;
                }
                break;
            }
        }
        //cout << " - " << readString << "\n";
    }

    //Reserved space for loading-based error trapping.

    //<ROM Load> not found
    if (romLoad[0] == NULL)
    {
        message("No <ROM Load> file specified in leveldata.txt.");
        return 1;
    }

    //<ROM Save> not found
    if (romSave[0] == NULL)
    {
        message("No <ROM Save> file specified in leveldata.txt.");
        return 1;
    }

    //<Map File> not found
    if (map[0] == NULL)
    {
        message("No <Map File> specified in leveldata.txt.");
        return 1;
    }

    //<Name> not found
    if (levelName[0] == NULL)
    {
        message("No level <Name> specified in leveldata.txt.");
        return 1;
    }

    //<Tileset> not found
    if (tileset == -1)
    {
        message("No <Tileset> specified in leveldata.txt.");
        return 1;
    }

    //Tileset illegal
    if ((tileset < 0) || (tileset > 7) || (tileset == 5))
    {
        cout << "The tileset type specified in leveldata.txt is illegal.\n";
        message("The tileset value must be 0-4, 6 or 7.");
        return 1;
    }

    //Illegal level number
    if ((level < 0) || (level > 29))
    {
        message("The level number specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal rating
    if ((rating < 0) || (rating > 3))
    {
        message("The rating specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal number of lemmings
    if ((lemmings != 1) && (lemmings != 2) && (lemmings != 4) && (lemmings != 5) && (lemmings != 10) && (lemmings != 20))
    {
        cout << "The number of lemmings specified in leveldata.txt is illegal.\n";
        message("Supported values are 1, 2, 4, 5, 10 or 20.");
        return 1;
    }

    //Illegal to be saved
    if ((toBeSaved < 0) || (toBeSaved > 100))
    {
        message("The percentage of lemmings to be saved specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal release rate

    if ((releaseRate < 0) || (releaseRate > 99))
    {
        message("The release rate specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal time
    if ((timeMinutes < 0) || (timeMinutes > 9))
    {
        message("The time limit specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal climbers
    if ((climbers < 0) || (climbers > 99))
    {
        message("The climber value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal floaters
    if ((floaters < 0) || (floaters > 99))
    {
        message("The floater value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal bombers
    if ((bombers < 0) || (bombers > 99))
    {
        message("The bomber value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal blockers
    if ((blockers < 0) || (blockers > 99))
    {
        message("The blocker value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal builders
    if ((builders < 0) || (builders > 99))
    {
        message("The builder value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal bashers
    if ((bashers < 0) || (bashers > 99))
    {
        message("The basher value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal miners
    if ((miners < 0) || (miners > 99))
    {
        message("The miner value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal diggers
    if ((diggers < 0) || (diggers > 99))
    {
        message("The digger value specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal fall distance
    if ((fallDistance < 0) || (fallDistance > 255))
    {
        message("The fall distance specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal trap type
    if ((trapType < 0) || (trapType > 4))
    {
        message("The trap type specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal trap X position
    if ((trapX < 0) || (trapX > 896))
    {
        message("The X position of the trap specified in leveldata.txt is illegal.");
        return 1;
    }

    //Illegal trap Y position
    if ((trapY < 0) || (trapY > 152))
    {
        message("The Y position of the trap specified in leveldata.txt is illegal.");
        return 1;
    }

    /*
    STAGE TWO: Convert the map format to one that the game will understand.
    */

    switch(mapFileType)
    {
        //Master System Lemmings Map (*.MLM)
        case 1:
        {
            mapSize = SMSMap(map);
            if (mapSize == -1)
                {
                    message("The map file failed to load properly.");
                    return 1;
                }
            break;
        }

        //Mappy (*.MAP)
        case 2:
        {
            mapSize = MappyMap(map);
            if (mapSize == -1)
                {
                    message("The map file failed to load properly.");
                    return 1;
                }
            break;
        }
        default:
        {
            cout << mapFileType << "\n";
            message("The map file type is not supported.");
            return 1;
        }
    }


    /*
    STAGE THREE: Modify the ROM
    */

    //Load the ROM into memory
    romLoad_ptr = fopen (romLoad, "rb");
    num = fread(romData, 1, 262144, romLoad_ptr);
    if (num != 262144)
    {
        message ("The ROM file failed to load.");
        for (i=0; i<num; i++)
        {
            cout << hex << (int)romData[i] << " ";
        }
        return 1;
    }
    fclose(romLoad_ptr);

    //Edit the ROM data
    int offset = 260009 + (((rating*30)+level)*17);
    romData[(offset++)] = (unsigned char) lemmings;
    romData[(offset++)] = (unsigned char) toBeSaved;
    romData[(offset++)] = (unsigned char) releaseRate;
    romData[(offset++)] = (unsigned char) timeMinutes;
    romData[(offset++)] = (unsigned char) climbers;
    romData[(offset++)] = (unsigned char) floaters;
    romData[(offset++)] = (unsigned char) bombers;
    romData[(offset++)] = (unsigned char) blockers;
    romData[(offset++)] = (unsigned char) builders;
    romData[(offset++)] = (unsigned char) bashers;
    romData[(offset++)] = (unsigned char) miners;
    romData[(offset++)] = (unsigned char) diggers;

    romData[(offset++)] = (unsigned char) 100/lemmings;

    //Map pointer, which is where the map data will be stored in the ROM. It writes 0x7670 in little endian order.
    romData[(offset++)] = (unsigned char) 0x70;
    romData[(offset++)] = (unsigned char) 0x76;

    //Level name pointer, which is where the game searches for the level's name during runtime. The address written will be 0xBFA2.
    romData[(offset++)] = (unsigned char) 0xa2;
    romData[(offset++)] = (unsigned char) 0xbf;

    //Move the offset to 0x3FFA2, which is where the level name data will actually be stored in the ROM.
    offset = 0x3ffa2;
    i=0;

    //Write the level name into the memory block
    while (levelName[i] != '\0')
    {
        romData[offset++] = (unsigned char) levelName[i++];
    }

    //Write the map data into the ROM. This data will be stored at 0x7670.
    offset = 0x7670;

    //Store the level tileset, eg. grass, fire, ice, sand...
    romData[(offset++)] = (unsigned char) tileset;

    //Store the ROM page the level will be stored in the ROM.
    romData[(offset++)] = (unsigned char) 0x0e;

    //Store the address of the level within the ROM page. Little endian.
    romData[(offset++)] = (unsigned char) 0x20;
    romData[(offset++)] = (unsigned char) 0xb0;

    //The fall distance is attributed to the map, rather than the level.
    romData[(offset++)] = (unsigned char) fallDistance;

    //Go to the ROM location where the actual map data will be stored.
    offset =  0x3b020;

    //Inject the map into the ROM.
    for (int i=0; i<mapSize; i++)
    {
        romData[offset++] = (unsigned char) mapFormat[i];
    }

    //Implement the level's trap.
    ImplementTrap(trapType, trapX, trapY, level, rating);

    //Now to save the edited ROM.
    romSave_ptr = fopen(romSave, "wb");
    num = fwrite (romData, 1, 262144, romSave_ptr);
    if (num == 262144)
    {
        //cout << "The ROM saved successfully!";
        message("The ROM saved successfully!");
    }
    else
    {
        message("The ROM failed to save.");
    }
    fclose(romSave_ptr);
    return 0;
}
